// GLSL Vertex Shader for GPU based Volume Raycasting

#version 110



////////////////////////////////////////////////////////////////////////
/// 
/// Gradient Code 
///
////////////////////////////////////////////////////////////////////////


uniform vec3 dX;
uniform vec3 dY;
uniform vec3 dZ;

vec3 gradientCalc(vec3 pos, sampler3D tex, sampler1D transferFunction)
{
	// calc gradient in volume
	vec3 sample1;
	vec3 sample2;
	sample1.x = texture3D(tex, pos + dX).a;
	sample2.x = texture3D(tex, pos - dX).a;
	sample1.y = texture3D(tex, pos + dY).a;
	sample2.y = texture3D(tex, pos - dY).a;
	sample1.z = texture3D(tex, pos + dZ).a;
	sample2.z = texture3D(tex, pos - dZ).a;
	
	// scale with transfer function
	vec3 alpha1;
	vec3 alpha2;
	alpha1.x = texture1D(transferFunction, sample1.x).a;
	alpha2.x = texture1D(transferFunction, sample2.x).a;
	alpha1.y = texture1D(transferFunction, sample1.y).a;
	alpha2.y = texture1D(transferFunction, sample2.y).a;
	alpha1.z = texture1D(transferFunction, sample1.z).a;
	alpha2.z = texture1D(transferFunction, sample2.z).a;
	
	// central difference and normalization
	return normalize(alpha2 - alpha1);
}


vec3 gradientCalcIso(vec3 pos, sampler3D tex)
{
	// calc gradient in volume
	vec3 sample1;
	vec3 sample2;
	sample1.x = texture3D(tex, pos + dX).a;
	sample2.x = texture3D(tex, pos - dX).a;
	sample1.y = texture3D(tex, pos + dY).a;
	sample2.y = texture3D(tex, pos - dY).a;
	sample1.z = texture3D(tex, pos + dZ).a;
	sample2.z = texture3D(tex, pos - dZ).a;
	
	// central difference and normalization
	return normalize(sample2 - sample1);
}


////////////////////////////////////////////////////////////////////////
///
/// Shading
/// 
////////////////////////////////////////////////////////////////////////


// #####################################################################
// 
// Blinn Phong 
// 
// #####################################################################

// Blinn Phong Illumination
// N = normalized normal vector
// V = normalized (eyePosition - pos)
// L = normalized (lightPosition - pos)

vec3 blinnPhongIllumination(vec3 N, vec3 V, vec3 L, vec3 dir)
{
	// material properties
	vec3 Ka = vec3(0.1, 0.1, 0.1); // ambient
	vec3 Kd = vec3(0.5, 0.5, 0.5); // diffuse
	vec3 Ks = vec3(0.2, 0.2, 0.2); // specular
	float shininess = 80.0;
	
	// light properties
	vec3 lightColor = vec3(1.0, 1.0, 1.0);
	vec3 ambientLight = vec3(0.3, 0.3, 0.3);
	
	// Calculate halfway vector
	 vec3 H = normalize(L - dir);
	// Calculate halfway vector
	//vec3 H = normalize(L + V);
	
	// Compute ambient erm
	vec3 ambient = Ka * ambientLight;
	
	// Compute the diffuse term
	float diffuseLight = max(dot(L, N), 0.0);
	vec3 diffuse = Kd * lightColor * diffuseLight;
	
	// Compute the specular term
	float specularLight = pow(max(dot(H, N), 0.0), shininess);
	vec3 specular = Ks * lightColor * specularLight;
	
	return ambient + diffuse + specular;
}


uniform vec3       eye;
								   
vec4 phongShading(vec3 pos, vec3 lightPos, vec3 dir, sampler3D tex, sampler1D transferFunction)							   
{

	
	float sample = texture3D(tex, pos).a;
	vec4 result = texture1D(transferFunction, sample);
	if (result.a > 0.01)
	{
		vec3 N = gradientCalc(pos, tex, transferFunction);
		// calculate light and viewing direction
		vec3 V = normalize(eye - pos);
		vec3 L = normalize(lightPos - pos);
		// add local illumination
		result.rgb += blinnPhongIllumination(N, V, L, dir);
	}
	return result;
}

vec4 phongShadingIso(vec4 color, vec3 pos, vec3 dir, vec3 lightPos,  sampler3D tex)							   
{

	
	float sample = texture3D(tex, pos).x;
	vec4 result = color;
	if (result.a > 0.01)
	{
		vec3 N = gradientCalcIso(pos, tex);
		// calculate light and viewing direction
		vec3 V = normalize(eye - pos);
		vec3 L = normalize(lightPos - pos);
		// add local illumination
		result.rgb += blinnPhongIllumination(N, V, L, dir);
	}
	
	return result;
}











///////////////////////////////////////////////////////////////////////
///
/// Raycasting
///
///////////////////////////////////////////////////////////////////////

struct Ray
{
	vec3 rayPos;
	vec3 rayDir;
	vec3 rayStep;
	float rayLength;
	float rayStepLength;
	float length_acc;
};


vec4 raycastIsoPath(in    vec4 color,
					in    vec4 isoColor,
					inout Ray ray, 
					in    int max_samples,
					in    sampler3D vol, 
					in    float iso_value,
					in    vec3 lightPos)
{

	vec4 dst = color;
	vec4 scalar = vec4(0.0);
	vec3 rayStep = ray.rayStep;
	float rayStepLength = ray.rayStepLength;
	bool stepDecreased = false;

	for(int i = 0; i < max_samples; ++i)
	{
		scalar = texture3D(vol, ray.rayPos);
		if (ray.length_acc > 0.0 && ray.length_acc < 0.99 * ray.rayLength && scalar.a > iso_value)
		{
			// refinement as proposed by Hadwiger2005EG
			vec3 xnew = ray.rayPos, xnear, xfar;
			vec4 fnew, fnear, ffar;
			xnear = ray.rayPos-rayStep;
			xfar  = ray.rayPos;			
			for (int j = 0; j < 5; j++)
			{
				fnear = texture3D(vol, xnear);
				ffar  = texture3D(vol, xfar);
				
				xnew = (xfar - xnear) * (iso_value - fnear.a) / (ffar.a - fnear.a) + xnear;
				fnew = texture3D(vol, xnew);
				if (fnew.a > iso_value) xnear = xnew; else xfar = xnew;
			}
			dst = phongShadingIso(isoColor, xnear, ray.rayDir, lightPos, vol);
			break;
		}		
		if (scalar.a > iso_value-0.1 && !stepDecreased) 
		{
			rayStep = rayStep * 0.2;
			rayStepLength = length(rayStep);
			stepDecreased = true;
		}

		if (scalar.a < iso_value-0.1 && stepDecreased) 
		{
			rayStep = rayStep * 5.0;
			rayStepLength = length(rayStep);
			stepDecreased = false;
		}
	
		ray.rayPos += rayStep;
		ray.length_acc += rayStepLength;
		
		if(ray.length_acc >= 0.99 * ray.rayLength)
		{
			break;
		}
	}
	return dst;
}

vec4 raycastWithShadingPath(in    vec4 color, 
						inout Ray ray, 
						in    int max_samples,
						in    sampler3D vol, 
						in    sampler1D transfer_func,
						in    vec3 lightPos)
{
	float scalar = 0.0;
	vec4 dst = color;	
	vec4 src = vec4(0.0);

	for(int i = 0; i < max_samples; ++i)
	{
		src = phongShading(ray.rayPos, lightPos,ray.rayDir, vol, transfer_func);
		// post multiply alpha for opacity correction
		src.rgb *= src.a;
		// compute volume rendering integral
		dst = (1.0 - dst.a) * src  + dst;
		ray.rayPos += ray.rayStep;
		ray.length_acc += ray.rayStepLength;
		
		if(ray.length_acc >= 0.99 * ray.rayLength || dst.a > 0.99)
		{
			break;
		}
	}
	return dst;
}



vec4 raycastPath(in    vec4 color, 
				 inout Ray ray,
				 in    int max_samples,
				 in    sampler3D vol,
				 in    sampler1D transfer_func)
{
	float scalar = 0.0;
	vec4 dst = color;	
	vec4  src = vec4(0.0);

	for(int i = 0; i < max_samples; ++i)
	{
		// read value from 3D volume
		scalar = texture3D(vol, ray.rayPos).a;
		// apply transfer function
		src = texture1D(transfer_func, scalar);
		// post multiply alpha for opacity correction
		src.rgb *= src.a;
		// compute volume rendering integral
		dst = (1.0 - dst.a) * src  + dst;
		ray.rayPos += ray.rayStep;
		ray.length_acc += ray.rayStepLength;
		
		if(ray.length_acc >= 0.99 * ray.rayLength || dst.a > 0.99)
		{
			break;
		}
	}
	return dst;
}


vec4 preintRaycastPath(in    vec4 color, 
					   inout Ray  ray,
					   in    int max_samples,				// maximum number of samples along ray
					   in    sampler3D vol,
					   in    sampler2D preint_func)
{

	vec2 scalar = vec2(0.0);
	vec4  dst  = color;
	vec4  src = vec4(0.0);

	for(int i = 0; i < max_samples; ++i)
	{
		// read value from 3D volume
		scalar.y = texture3D(vol, ray.rayPos).a;
		// apply preint transfer function
		src = texture2D(preint_func, scalar);
		// compute volume rendering integral
		dst = (1.0 - dst.a) * src  + dst;
		scalar.x = scalar.y;
		ray.rayPos += ray.rayStep;
		ray.length_acc += ray.rayStepLength;
		
		if(ray.length_acc >= 0.99 * ray.rayLength || dst.a > 0.99)
		{
			break;
		}
	}
	return dst;
	
}


Ray initRay(in vec3 rayStart, 
			in vec3 rayDir, 
			in vec3 scaleFactors, 
			in float delta)
{
	Ray ray;

	ray.rayPos = rayStart.xyz * scaleFactors;
	ray.rayDir = rayDir.xyz * scaleFactors;
	ray.rayLength = length(ray.rayDir);
    ray.rayStep = normalize(ray.rayDir) * delta;
    ray.rayStepLength = length(ray.rayStep);
	ray.length_acc = 0.0;

	return ray;
}

// vertex to fragment shader variables

// mandatory variables
uniform sampler1D  transfer_function;
uniform sampler2D  preint_function;
uniform sampler2D  ray_start;
uniform sampler2D  ray_dirTex;
uniform sampler2D  jitter;
uniform sampler3D  volume;
uniform vec3	   texScaleFactors;


// user parameters
uniform float 	   iso_value;
uniform float      sample_distance;
uniform int        max_samples;
uniform vec2	   tileSize;
uniform bool	   useShading; 
uniform bool	   useJitter;
uniform bool	   usePreint;
uniform vec3	   lightPos;

void main()
{
	// lookup ray start
	vec3 rayStart = vec3(texture2D(ray_start, gl_TexCoord[0].st));
	// lookup ray direction and length
	vec4 rayDirAndLength = texture2D(ray_dirTex, gl_TexCoord[0].st);

	// discard fragments with zero ray length
	if(rayDirAndLength.w == 0.0) 
		discard;

	Ray ray = initRay(rayStart, vec3(rayDirAndLength), texScaleFactors, sample_distance);
	
	
	
	
	
	
	vec4 res = vec4(0.0); 
	
	if(useJitter)
	{
		float offset =  texture2D(jitter, gl_TexCoord[0].st * tileSize.xy).x ;
		ray.rayPos += offset * ray.rayStep;
		ray.length_acc = length(offset * ray.rayStep);
	}

	

	

	// Volume Rendering Paths
	// Volume Rendering with pre-integrated Transfer Functions
	if(usePreint)
		res = preintRaycastPath(res, ray, max_samples, volume, preint_function);
	else
	{
		// Volume Rendering with local illumation based on centralized gradients
		if(useShading)
			res = raycastWithShadingPath(res, ray, max_samples, volume, transfer_function, lightPos);
		else
		// Basic volume rendering with 1D transfer functions
			res = raycastPath(res, ray, max_samples, volume, transfer_function);
	}
	
	
	
	gl_FragColor = res;

}